from pystencils.backends.cbackend import generate_c, get_headers
from pystencils.astnodes import KernelFunction
from pystencils.enums import Backend
from jinja2 import Environment, PackageLoader, StrictUndefined

from pystencils_benchmark.enums import Compiler
from pathlib import Path

_env = Environment(loader=PackageLoader('pystencils_benchmark'),
                   undefined=StrictUndefined,
                   keep_trailing_newline=True,
                   trim_blocks=True, lstrip_blocks=True)


def _kernel_header(kernel_ast: KernelFunction,
                   dialect: Backend = Backend.C,
                   *,
                   template_file: str,
                   additional_jinja_context: dict = None) -> str:

    function_signature = generate_c(kernel_ast, dialect=dialect, signature_only=True)
    header_guard = f'_{kernel_ast.function_name.upper()}_H'

    jinja_context = {
        'header_guard': header_guard,
        'function_signature': function_signature,
    }
    if additional_jinja_context is not None:
        jinja_context.update(additional_jinja_context)

    header = _env.get_template(template_file).render(**jinja_context)
    return header


def _kernel_source(kernel_ast: KernelFunction,
                   dialect: Backend = Backend.C,
                   *,
                   template_file: str,
                   additional_jinja_context: dict = None) -> str:

    kernel_name = kernel_ast.function_name
    function_source = generate_c(kernel_ast, dialect=dialect)
    headers = {f'"{kernel_name}.h"', '<math.h>', '<stdint.h>'}
    headers.update(get_headers(kernel_ast))

    jinja_context = {
        'function_source': function_source,
        'headers': sorted(headers),
        'timing': True,
    }

    if additional_jinja_context is not None:
        jinja_context.update(additional_jinja_context)

    source = _env.get_template(template_file).render(**jinja_context)
    return source


def compiler_toolchain(path: Path, compiler: Compiler, likwid: bool) -> None:
    name = compiler.name
    jinja_context = {
        'compiler': name,
        'likwid': likwid,
    }

    files = ['Makefile', f'{name}.mk']
    for file_name in files:
        with open(path / file_name, 'w+') as f:
            template = _env.get_template(file_name).render(**jinja_context)
            f.write(template)


def copy_static_files(path: Path, *, source_file_suffix='.c') -> None:
    src_path = path / 'src'
    src_path.mkdir(parents=True, exist_ok=True)
    include_path = path / 'include'
    include_path.mkdir(parents=True, exist_ok=True)

    files = ['timing.h', 'timing.c']
    for file_name in files:
        template = _env.get_template(file_name).render()
        if file_name[-1] == 'h':
            target_path = include_path / file_name
        elif file_name[-1] == 'c':
            target_path = src_path / file_name
            target_path = target_path.with_suffix(source_file_suffix)
        else:
            target_path = path / file_name
        with open(target_path, 'w+') as f:
            f.write(template)


def setup_directories(path: Path):
    if path is None:
        path = Path('.')
    else:
        path.mkdir(parents=True, exist_ok=True)
    src_path = path / 'src'
    src_path.mkdir(parents=True, exist_ok=True)
    include_path = path / 'include'
    include_path.mkdir(parents=True, exist_ok=True)
    return src_path, include_path
